import audiobusio
import board

from ulab import numpy as np
from ulab.scipy.signal import spectrogram
import array

#---| Mic Configuration |---------------------------
SAMPLERATE = 16000
SAMPS = fft_size = 128
# there seems to be a difference of 50hz when calculating from Pure Tones
# so we'll cheat & just add 50 to the "listened" tone
DIFF = 50 

# ============== make the PDM mic object =============== #

class PDM():

    def __init__(self):
        self.mic = audiobusio.PDMIn(
                    board.GP27_A1,
                    board.GP0,
                    sample_rate=SAMPLERATE,
                    bit_depth=16,
                    startup_delay=0.05,
                    mono=True
                    )
        self.give = 30 # +/- "give" when looking fr an Hz in self.getNote() 
        self.samples_bit = array.array('H', [0] * (fft_size+3))
        self.listen()
        return
    
    def setGive(self, g):
        # change the allowance / give for accepting an Hz
        self.give = g

    def listen(self):
        self.mic.record(self.samples_bit, len(self.samples_bit))
        samples = np.array(self.samples_bit[3:])
        spectrogram1 = spectrogram(samples)
        # spectrum() is always nonnegative, but add a tiny value
        # to change any zeros to nonzero numbers
        spectrogram1 = np.log(spectrogram1 + 1e-7)
        spectrogram1 = spectrogram1[1:(fft_size//2)-1]
        self.peak_idx = np.argmax(spectrogram1)
        self.peak_freq = (self.peak_idx * SAMPLERATE / SAMPS / 2) + DIFF #  Add the DIFF cludge to the result!
        return

    def isNote(self, expectedNote):
        self.listen()
        # check to see if the "heard note" ( playedNote )
        # is the same as the one expected ( expectedNote ) 
        # also within the range of "give"
        heardNote = self.whatNote()
        if expectedNote - self.give < heardNote < expectedNote + self.give:
            return(True)
        else:
            return(False)

    def whatNote(self):
        #self.listen()
        return(self.peak_freq)


#make a PDM mic object
pdm_mic = PDM()
