import time
import config
import displayio
import random
# vectorio is now part of CircuitPython (so no need for external library)
# but we still need to import it
from vectorio import Polygon, Rectangle, Circle
from adafruit_display_text import label
import utilities

screen = config.screen
display = config.display

WIDTH = HEIGHT = 240
CENT_XY = (int(WIDTH/2), int(HEIGHT/2))
HEX = 'FEDCBA9876543210'
# a cludge to create a screen wipe to remove ghost images from game pieces
BLACKPALETTE = displayio.Palette(1)
BLACKPALETTE[0] = 0x000
WIPE = Rectangle(pixel_shader=BLACKPALETTE,
                 width=WIDTH, height=HEIGHT, x=0, y=0)
WIPEGROUP = displayio.Group()
WIPEGROUP.append(WIPE)

# Create the text label to display INFO
INFO = label.Label(config.font, text="", color=0x00FFAA, scale=4)
INFO.x, INFO.y = -1000, 120
WIPEGROUP.append(INFO)
# and add the WIPE
screen.append(WIPEGROUP)
# refresh the screen
display.show(screen)


# some functions for setting up asteroids
def getRandsUpTo(tot, divs, r, r2):
    ''' make an array of random numbers "sorted" (small to large) for points around asteroid'''
    incs = int(tot / divs)
    ranges = [0]
    rots = []
    lens = []
    # cycle up to maximum range
    for i in range(divs):
        #add to ranges i+1 * degree incs
        ranges.append((i+1)*incs)
        # make up a random rotation(degree) between ranges[i..i+i]
        rots.append(random.randint(ranges[i], ranges[i+1]))
        # make up a random length for this node
        lens.append(random.randint(r2, r))
    return(rots, lens)


def setRandPolyNodes(level, r, r2):
    ''' initial set up of a node dict for an asteroid 
    nade_number : rotation & length from centre
    self.nodeDict = {x: {'rot' : x, 'len' : y}...} '''
    # calculate how many points the polygone will have depending on level
    numPoints = [7, 5, 3, 3][level]
    degList, lenList = getRandsUpTo(360, numPoints, r, r2)
    #set up a blank dictionary
    tempDict = {}
    for i, d in enumerate(degList):
        #cycle through vals of degList & make new Dict Keys & pass rotation & len as sub keys/values
        tempDict[i] = {'rot': d, 'len': lenList[i], 'xy': (0, 0)}
    return(tempDict)
## =========================


class Game():
    ''' main class to control all game elements
    e.g. player, bullets, asteroids, score, level etc'''

    pieces = []
    player = None  # game object devines player object
    # the Game has an imGroup we can add all images to...
    imGroup = displayio.Group()
    # and add the  Game.imGroup to the screen
    screen.append(imGroup)
    # make the images draw / update
    display.show(screen)

    def testShowObs(msg="BEFORE"):
        print("vectorIOs in Game.imGroup = {} {}".format(len(Game.imGroup), msg))
        if Game.getObjects(Bullet):
            print(Game.getObjects(Bullet))
        else:
            print("no bullets onscreen \n images now {}".format(
                Game.getObjects(Bullet)))

    # pop a game piece (can pass a list of objects or single object)
    def popPiece(who):
        # cludge screen wipe to delete ghosts
        # WIPE.x = 0 brings it to the
        WIPE.x = 0
        display.show(screen)
        #Game.testShowObs()
        '''if type(who) == list:  # multi delete of image objects
            for o in who:
                Game.imGroup.remove(o.shape)
                Game.pieces.remove(o)
                del who
        else:''' # old list of objects to delete (unused now)
            # print(who.shape)
        Game.imGroup.remove(who.shape)
        if type(who) == Asteroid:
            Game.imGroup.remove(who.txtGroup) # take the hit counters out too
        Game.pieces.remove(who)
        #Game.testShowObs("AFTER")
        WIPE.x = -WIDTH
        display.show(screen)
        return

        # update the screen
        #display.refresh()
        #display.show(screen)

    # adda n object (who) to the Game pieces
    def addPiece(who):
        Game.pieces.append(who)
        Game.imGroup.append(who.shape)
        # check on specific game pieces
        if type(who) == Asteroid:
            Game.imGroup.append(who.txtGroup)
        elif type(who) == Player:
            Game.player = who
        return

    # cycle through the Game's pieces list & return only "what" objects
    def getObjects(what):
        return([ob for ob in Game.pieces if type(ob) == what])

    def getPlayer():
        return(Game.player)

    # a hit occurred when a bullet hit an asteroid
    def bulletHitTarget(bullet, target):
        # print("bullet {} hit asteroid {}".format(bullet, target))
        # send the hit to the target (Asteroid) and the Bullet
        bullet.hit()  # get rid of the bullet
        target.hit(bullet)  # destroy & split (if big enough) Asteroid
        if Game.getObjects(Asteroid) == []: # no asteroids left
            Game.endGame("WIN", 0x00AA22, 10)
            
    def updateBulletTargets(who):
        for b in Game.getObjects(Bullet):
            b.addTarget(who)        
        
    def setInfo(msg="", col=0x00FFAA, zzz=1, xpos=30, ypos=30):
        # quick clunky text display for events
        INFO.color = col
        INFO.text = msg
        if xpos != 30:
            xpos = INFO.width
        INFO.x = int(xpos)
        INFO.y = int(ypos)
        display.show(screen)
        # cludgy pause
        time.sleep(zzz)
        INFO.x = -240 # send offscreen again      
        
    def endGame(msg="DEAD!", col=0xFF1111, zzz=10):
        # pop all the game pieces out
        for o in Game.pieces:
            Game.popPiece(o)
        # call the info function to show end game info
        Game.setInfo(msg, col, zzz)
        return

    # =================================================================================== game object
    # INIT the Game object
    def __init__(self):
        print("oi oi! wake up ! game on!")
        self.level = 0
        #self.state = False
        self.player = Player()
        self.spawnAsteroids(WIDTH/2, HEIGHT/2)
        self.firePause = 1
        self.firePauseCount = 0
        #print("Game starts with {} pieces".format(Game.pieces))
        self.state = True
        #display.show(screen)
        Game.setInfo("GAME\nON!", 0x00FF44, 1, 0, 30)
        return

    def spawnAsteroids(self, x, y):
        Asteroid(x, y)
        # Game.addPiece(a)
        return

    def fire(self):
        # we now create & add a Bullet() oject to the Game.pieces list
        if self.firePauseCount == 0:
            Game.addPiece(Bullet())
            self.firePauseCount = 0.1

    def rotate(self, d=1):
        #rotate the player
        self.player.rotate(0.1 * d)

    def boost(self):
        #print("game boosts player")
        self.player.boost()

    # main game loop (moves all peices / checks for collisions)
    def loop(self):
        # like an exit frame?
        for o in Game.pieces:
            #  cycle through all the Game.pieces & move them
            o.move()
            self.checkCollision()

        # clunky way of creating a pause between shots
        if self.firePauseCount != 0:
            self.firePauseCount += 0.1
            if self.firePauseCount >= self.firePause:
                self.firePauseCount = 0
    # game object checks for Asteroid / Player collissions
    def checkCollision(self):
        # check to see if an asteroid has hit the player
        #if self.state:
        # if the player is immune (recently hit or safe pill on)
        if not self.player.immune:
            for asteroid in Game.getObjects(Asteroid):
                if utilities.isInside(asteroid.x, asteroid.y, asteroid.r, self.player.x, self.player.y):
                    Game.setInfo("CRASH!", 0xFF5500, 0.2)
                    self.player.kill()


############################## Pieces Classes ##################################
################ ========== Poly(gon) Class =========== ########################
class Poly():
    ''' Base Class for all polygons in this program '''

    def __init__(self, x=int(WIDTH/2), y=int(HEIGHT/2), level=0, c=0xFFFFFF):
        # all polygons work around "(0,0)" coordinates to start with
        self.x = self.y = 0
        self.hits = 1
        # movement / vector / rotation properties / scale
        self.vect = (0, 0)
        self.rot = 0  # rotation of object
        self.rotSpeed = 0  # speed of rotation
        self.d = 0  # direction
        self.sf = 1.0
        self.vel = 1  # velocity
        # colour palette info for this object + save oc (original color) =================
        self.c = self.oc = c
        self.col = displayio.Palette(1)
        self.col[0] = self.c
        # image shape / group info ===============================
        self.shape = Polygon(pixel_shader=self.col,
                             # set the minimum points to (0,0)x3 as a holder
                             points=[(0, 0), (0, 0), (0, 0)],
                             x=0, y=0)  # x & y must be 0 to start with ("moved" later with setNewXY(x,y))

    def calcNewShape(self):
        '''
        we need to re-calculate the points list for this polygon
        we create a blank array & cycle through all the nodes in the polygon
        with the polyDict looking for the angle + len (and add current rotation & scale)
        '''
        newShape = []
        for k, v in self.polyDict.items():
            newLoc = utilities.calcXY(
                v['rot'] + self.rot,
                v['len']*self.sf  # sf = scale factor
            )
            # we now calculat the new location for each point
            # by adding the calculated rotation + radius x,y tuple
            # to the x,y tuple of the overall object
            newLoc = utilities.addTuples(
                (int(newLoc[0]), int(newLoc[1])), (int(self.x), int(self.y))
            )
            v['xy'] = newLoc
            newShape.append(newLoc)
        self.shape.points = newShape
        return(newShape)

    # Poly.move()
    def move(self):
        # double check to see if this poly is "alive"(!)
        if self.hits > 0:
            # check to see if the velocity is not 0
            if self.vel != 0:
                self.x += self.vect[0] * self.vel
                self.y += self.vect[1] * self.vel

            # check if x is outside the width of the screen
            if self.x < -self.r:
                self.x = WIDTH+self.r
            if self.x > WIDTH+self.r:
                self.x = -self.r
            # check if y is outside the height of the screen
            if self.y < -self.r:
                self.y = HEIGHT+self.r
            if self.y > HEIGHT+self.r:
                self.y = -self.r

            # keep total rotation for this object(needed?)
            self.rot += self.rotSpeed
            # redraw the shape rotated
            self.shape.points = self.calcNewShape()
            return
        else: # dead!
            self.x = -1000 # offscreen cludge

    # ============ Some Getters for info / returns

    def getDict(self):
        for k, v in self.polyDict.items():
            print("v={} xy={} for {}".format(v, (self.x, self.y), self))
        print("polyPoints = {}".format(self.polyPoints))

    def getInfo(self):
        print("rot={}\nvel={}\nvect={}".format(self.rot, self.vel, self.vect))

    def getVect(self):
        return(self.vect)

    # ============ Some setters for changing rotation speed / vector etc

    def setRotSpeed(self, r):
        # clamp the new rotation speed to within -3 & 3
        self.rotSpeed = utilities.clamp(r, -5, 5)
        if self.rotSpeed > 0:
            #self.d = direction (so we can convert to clockwise/anti)
            self.d = 1
        elif self.rotSpeed < 0:
            self.d = -1
        return

    def setVect(self, ang):
        self.vect = utilities.getVect(ang)

    def setRot(self, ang):
        self.rot = ang

    def setNewXY(self, x, y):
        # set the x, y values of this object to a new x,y
        self.x = x
        self.y = y
        return

    def setCol(self, c):
        self.c = c
        self.col[0] = self.c
        return

    def scale(self, sf=1.0):
        #change the scale factor
        if sf >= 0.1 and sf <= 5.0:
            self.sf = sf
        else:
            print("scale factor outide params (0.1 to 5.0)")
        self.calcNewShape()

    def setVel(self, v):
        self.vel = v
        
    #Poly setAll()
    def setAll(self, rotSpeed, vel, ang, x, y):
        self.setRotSpeed(rotSpeed)
        self.setVel(vel)
        self.setVect(ang)
        self.setNewXY(x,y)
        self.polyPoints = self.calcNewShape()
        return

########### ====== Asteroid Class = ####################################
''' Asteroid class is a sub of Poly with a randomly generated polygon'''


class Asteroid(Poly):
    ''' Asteroid Polygon
    hits = how many hits an asteroid can take depending on is level
    splits = how many sub asteroids it splits into 
    we've made 3 spits - but factored in future sizes / speeds
    '''
    rads = [
            random.randint(30, 40),
            random.randint(20, 29),
            random.randint(5, 10),
            random.randint(2, 5),
            random.randint(2, 3)
            ]
    speed = [
            utilities.getRand(1, 1.2),
            utilities.getRand(1.2, 1.4),
            utilities.getRand(2, 4),
            utilities.getRand(4, 7),
            utilities.getRand(7, 10)
            ]
    hits = [4,2,1]
    splits = [2,3,random.randint(3,6)]  
    rots = { # add these angles to the splits
        1: [random.randint(1,360)],
        2: [-90, 90], # 2 splits = -90 / 90 degrees
        3: [i * 120 for i in range(3)],
        4: [i * 90 for i in range(4)],
        5: [i * 72 for i in range(5)],
        6: [i * 60 for i in range(6)],
    }
    # Ateroid __init__
    def __init__(self, x=0, y=0, stage=0, ang=0):
        super().__init__()
        self.stage = stage
        self.r = Asteroid.rads[self.stage]
        self.r2 = int(self.r / 1.7)
        # how many hits to destroy this asteroid based on level
        self.hits = Asteroid.hits[self.stage]
        self.vel = Asteroid.speed[self.stage]
        self.polyDict = setRandPolyNodes(self.stage, self.r, self.r2)
        newCol = '0x{0}{0}{0}{0}{0}{0}'.format(random.randint(5, 9))
        hexInt = int(newCol, 16)
        self.oc = hexInt
        self.setCol(hexInt)
        
        if self.stage == 0:  # a brand new massive Asteroid (not split)
            ang = random.randint(1, 360)
            randXY = utilities.calcXY(ang, int(min(WIDTH, HEIGHT)/3.0))
            x, y = utilities.addTuples(randXY, CENT_XY)
            
        self.setAll(utilities.getRand(-1, 1), self.vel, ang, x,y)
        self.polyPoints = self.calcNewShape()

        self.txtGroup = displayio.Group()
        # Create the text label to display "hits left"
        self.txt = label.Label(config.font, text=str(self.hits), color=0x000)
        self.txt.x, self.txt.y = self.x, self.y
        self.txtGroup.append(self.txt)
        Game.updateBulletTargets(self)
        Game.addPiece(self)
        return
    
    def move(self):
        super().move() 
        # asteroids need to move thieir "hits" with the body
        self.txt.x, self.txt.y = int(self.x), int(self.y)
        
    # Asteroid.hit()
    def hit(self, b):
        # check to see if the stage of the Asteroid is splittable
        if self.stage < len(Asteroid.hits):
            if self.hits > 0:
                self.hits -= 1
                self.setCol(0xAA2200)
                self.move()
                self.setCol(self.oc)
                self.txt.text = str(self.hits)       
                Game.setInfo("POP!", 0x220000, 0.1, self.x, self.y)
                #print("I {} have {} hits left".format(self, self.hits))
                if self.hits == 0:
                    # Killed asteroid & within scope for splitting
                    # we now "split" the asteroid 
                    # based on stage of asteroid's life
                    if self.stage < len(Asteroid.hits)-1:
                        splits = Asteroid.splits[self.stage]
                        rotters = Asteroid.rots[Asteroid.splits[self.stage]]
                        for i in range(0, splits):
                            Asteroid(int(b.x),  # x pos
                                    int(b.y),  # y pos
                                    self.stage+1, # next stage of Asteroid (size/split etc)
                                    self.rot + rotters[i] # make the splits head in opposite directions
                                    )  
                        #print("I should be dead now {}".format(self))
                        #Game.testShowObs("BEFORE") # test call to see how many objects are live
                        Game.popPiece(self)  
                        Game.setInfo("BOOM!", 0x440000, 0.1, self.x, self.y)  
                                    
                        # Game.testShowObs("AFTER")  # test call to see how many objects are live AFTER pops
                    else:
                        # last death for Asteroid split
                        Game.popPiece(self)
                        del self

''' Player class is a sub Poly with a specific shape'''
class Player(Poly):
    
    # Player __init__
    def __init__(self):
        super().__init__()
        self.r = 5
        self.hits = 3
        self.immune = False
        self.vel = 0  # velocity
        self.mag = 0.05  # magnitude for decay
        # dictionary to create points for the arrow spaceship
        self.polyDict = {
            0: {'rot': 225, 'len': 10, 'xy': (0, 0)},
            1: {'rot': 0, 'len': 10, 'xy': (0, 0)},
            2: {'rot': 135, 'len': 10, 'xy': (0, 0)},
            3: {'rot': 150, 'len': 2, 'xy': (0, 0)},
        }
        self.centre()
        # add the Player to the Game CLass Variable "player"
        Game.addPiece(self)  # was Game.addPlayer(self)
        return

    def centre(self):
        cX, cY = CENT_XY  # extract the x,y from the CENT_XY tuple
        self.setNewXY(cX, cY)  # change this x,y position
        # creates point list & applies to Poly.shape.points
        self.polyPoints = self.calcNewShape()

    # Player.move()
    def move(self):
        super().move()
        ''' extra player move functions to decay rotation & velocity '''
        # velocity decay
        if self.vel > self.mag:
            self.vel = self.vel/1.1  # decay the velocety
            inject = HEX[int(self.vel)]
            newCol = '0xFF{0}{0}{0}{0}'.format(inject)
            hexInt = int(newCol, 16)
            self.changeCol(hexInt)
        else:
            self.vel = 0

        # rotation decay
        if self.rotSpeed >= self.mag or self.rotSpeed <= -self.mag:
            # decay the rotation
            self.setRotSpeed(self.rotSpeed - (self.mag * self.d))
            if self.vel != 0:
                self.vect = utilities.getVect(self.rot)
        else:
            self.rotSpeed = 0

    def changeCol(self, newCol):
        self.c = newCol
        self.col[0] = self.c

    def boost(self):
        ''' Sets the velocity of the player & increases as called (decreases in move())'''
        self.vect = utilities.getVect(self.rot)
        if self.vel == 0:
            self.vel = 1
        else:
            newVel = self.vel * 1.5
            self.vel = utilities.clamp(newVel, 0.1, 15)
        #self.move()
        return

    def rotate(self, speed):
        # create a temp value for speed of rotation by adding speed to self.rot
        s = self.rotSpeed + speed
        # make sure the speed is within clamp range
        self.rotSpeed = utilities.clamp(s, -3, 3)
        return

    def blink(self):
        # simple animation to re-spawn
        times = 4
        while times > 0:
            self.sf = 0.2
            self.move()
            time.sleep(0.1)
            self.sf = 1
            self.move()
            time.sleep(0.1)
            times -= 1
        # make the spaceship not immune to hits
        self.immune = False
        return

    def kill(self):
        # this is called from xxx
        print("kill called in Player")         
        self.immune = True
        self.hits -= 1
        Game.setInfo("LIVES {}".format(self.hits), 0xFF0022, 1) 
        self.setNewXY(int(WIDTH/2), int(HEIGHT/2))
        self.blink()
        if self.hits == 0:
            self.setAll(0,0,0,-240,0)
            Game.endGame()  # end the player
            
###################### ====== BULLET CLASS ======= #########################


class Bullet():

    def __init__(self):
        self.player = Game.getPlayer()
        self.x, self.y = self.player.x, self.player.y
        self.ox = self.x
        self.oy = self.y
        self.r = 2
        self.rot = self.player.rot
        self.vel = 1
        self.dist = 200  # pixels allowed to fire
        self.trav = 0  # pixels traveled
        self.immune = False
        # colour palette info for this object =================
        self.c = 0xFF0000  # red bullets!
        self.col = displayio.Palette(1)
        self.col[0] = self.c
        # image group container ===============================

        self.shape = Circle(pixel_shader=self.col,
                            radius=self.r,
                            x=int(self.x), y=int(self.y))
        # when this bullet is added to the Game pieces, let it know all the Ateroid objects
        self.targets = Game.getObjects(Asteroid)
        return

    # Bullet.move() # this move checks to see if this bullet hits a target (Asteroid)
    def move(self):
        # first move the bullet (if it's still "has power")
        if self.trav < self.dist:
            # calc the new x,y with original x,y + pythagorus radius / angle
            self.x, self.y = utilities.addTuples(
                (self.ox, self.oy),
                utilities.calcXY(self.rot, self.trav)
            )
            # if the bullet goes offscreen, erase the bullet
            if (self.x < 0 or self.x > WIDTH) or (self.y <0 or self.y > HEIGHT):
                Game.popPiece(self)
                return
            else:
                # bullet still onscreen, check for collisions
                self.shape.x = int(self.x)
                self.shape.y = int(self.y)
                self.trav += self.vel

                # check to see if this bullet has hit any asteroids (target)
                for target in self.targets:
                    if utilities.isInside(
                        # pass target x, y & radius
                        target.x,
                        target.y,
                        target.r2 + (target.r - target.r2)/2, # change the r2 for difficulty
                        # pass this bullet's x & y
                        self.shape.x,
                        self.shape.y
                    ):
                        # bullet is inside asteroid
                        # Game copes with all object removal
                        # print("bang! hit astroid {}".format(target))
                        self.targets.remove(target)
                        Game.bulletHitTarget(self, target)
        else:  # bullet now out of power
            self.hit("died")

    def levelComplete(self):
        # needs to go into game now
        pass

    # Bullet.hit()
    def hit(self, msg="taken out"):
        # print("Bullet {} {}".format(self, msg))
        try:
            Game.popPiece(self)
        except Exception as e:
            # bullet taken out in mass pop (game over)
            pass
        
    def addTarget(self, who):
        self.targets.append(who)


# ============= Make the Game Object ================ #
game = Game()
